package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.io.FileInputStream;
import java.io.FileOutputStream;

public final class FileStreams {
  /**
   * Mapping between FileInputStream objects and row index in T_INPUT_STREAM table
   */
  private static final ObjectRowIndexMap<FileInputStream> ourFileInputStream2RowIndex = new ObjectRowIndexMap<FileInputStream>();
  /**
   * Mapping between FileOutputStream objects and row index in T_OUTPUT_STREAM table
   */
  private static final ObjectRowIndexMap<FileOutputStream> ourFileOutputStream2RowIndex = new ObjectRowIndexMap<FileOutputStream>();

  private static final class InputStreamTable extends Table {
    public final StringColumn myFileName = new StringColumn("File Name");

    public InputStreamTable() {
      super("File Input Streams", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final InputStreamTable T_INPUT_STREAM = new InputStreamTable();

  private static final class FailedInputStreamTable extends Table {
    public final StringColumn myFileName = new StringColumn("File Name");

    public FailedInputStreamTable() {
      super("Failed File Input Streams", Table.MASK_FOR_POINT_EVENTS);
    }
  }
  private static final FailedInputStreamTable T_FAILED_INPUT_STREAM = new FailedInputStreamTable();

  private static final class ReadTable extends Table {
    private final ForeignKeyColumn myInputStream = new ForeignKeyColumn(T_INPUT_STREAM);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public ReadTable() {
      super("Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final ReadTable T_READ = new ReadTable();

  private static final class OutputStreamTable extends Table {
    private final StringColumn myFileName = new StringColumn("File Name");

    public OutputStreamTable() {
      super("File Output Streams", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final OutputStreamTable T_OUTPUT_STREAM = new OutputStreamTable();

  private static final class FailedOutputStreamTable extends Table {
    private final StringColumn myFileName = new StringColumn("File Name");

    public FailedOutputStreamTable() {
      super("Failed File Output Streams", Table.MASK_FOR_POINT_EVENTS);
    }
  }
  private static final FailedOutputStreamTable T_FAILED_OUTPUT_STREAM = new FailedOutputStreamTable();

  private static final class WriteTable extends Table {
    private final ForeignKeyColumn myOutputStream = new ForeignKeyColumn(T_OUTPUT_STREAM);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public WriteTable() {
      super("Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final WriteTable T_WRITE = new WriteTable();

  @MethodPattern("java.io.FileOutputStream:open(String)")
  public static final class FileOutputStream_open_Probe {
    public static void onReturn(
      final @This FileOutputStream stream,
      final @Param(1) String fileName
    ) {
      final int rowIndex = T_OUTPUT_STREAM.createRow();
      T_OUTPUT_STREAM.myFileName.setValue(rowIndex, fileName);
      ourFileOutputStream2RowIndex.put(stream, rowIndex);
    }

    public static void onUncaughtException(
      final @Param(1) String fileName
    ) {
      final int rowIndex = T_FAILED_OUTPUT_STREAM.createRow();
      T_FAILED_OUTPUT_STREAM.myFileName.setValue(rowIndex, fileName);
    }
  }

  @MethodPattern("java.io.FileOutputStream:close()")
  public static final class FileOutputStream_close_Probe {
    public static void onReturn(final @This FileOutputStream stream) {
      final int rowIndex = ourFileOutputStream2RowIndex.remove(stream);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      T_OUTPUT_STREAM.closeRow(rowIndex);
    }
  }

  @MethodPattern("java.io.FileOutputStream:writeBytes(byte[], int, int)")
  public static final class FileOutputStream_writeBytes_Probe {
    public static int onEnter(
      final @This FileOutputStream fileOutputStream,
      final @Param(3) int bytesWritten
    ) {
      return processWriteEnterImpl(fileOutputStream, bytesWritten);
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      processWriteReturnImpl(rowIndex);
    }
  }

  @MethodPattern("java.io.FileOutputStream:write(int)")
  public static final class FileOutputStream_write_Probe {
    public static int onEnter(final @This FileOutputStream fileOutputStream) {
      return processWriteEnterImpl(fileOutputStream, 1 /* one byte to be written */);
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      processWriteReturnImpl(rowIndex);
    }
  }

  private static int processWriteEnterImpl(final FileOutputStream fileOutputStream, final int bytesWritten) {
    final int streamRowIndex = ourFileOutputStream2RowIndex.get(fileOutputStream);
    if (streamRowIndex == Table.NO_ROW) {
      return -1;
    }

    // log individual write event
    final int writeRowIndex = T_WRITE.createRow();
    T_WRITE.myBytes.setValue(writeRowIndex, bytesWritten);
    T_WRITE.myOutputStream.setValue(writeRowIndex, streamRowIndex);
    return writeRowIndex;
  }

  private static void processWriteReturnImpl(final int rowIndex) {
    if (rowIndex != -1) {
      T_WRITE.closeRow(rowIndex);
    }
  }

  @MethodPattern("java.io.FileInputStream:open(String)")
  public static final class FileInputStream_open_Probe {
    public static void onReturn(
      final @This FileInputStream fileInputStream,
      final @Param(1) String fileName
    ) {
      final int rowIndex = T_INPUT_STREAM.createRow();
      T_INPUT_STREAM.myFileName.setValue(rowIndex, fileName);
      ourFileInputStream2RowIndex.put(fileInputStream, rowIndex);
    }

    public static void onUncaughtException(
      final @Param(1) String fileName
    ) {
      final int rowIndex = T_FAILED_INPUT_STREAM.createRow();
      T_FAILED_INPUT_STREAM.myFileName.setValue(rowIndex, fileName);
    }
  }

  @MethodPattern("java.io.FileInputStream:close()")
  public static final class FileInputStream_close_Probe {
    public static void onReturn(final @This FileInputStream fileInputStream) {
      final int rowIndex = ourFileInputStream2RowIndex.remove(fileInputStream);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      T_INPUT_STREAM.closeRow(rowIndex);
    }
  }

  @MethodPattern("java.io.FileInputStream:read()")
  public static final class FileInputStream_read_Probe {
    public static int onEnter(final @This FileInputStream fileInputStream) {
      return processReadEnterImpl(fileInputStream);
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue int returnValue
    ) {
      // non negative return value means that a byte has been successfully read
      final int readBytes = returnValue >= 0 ? 1 : 0;
      processReadReturnImpl(readRowIndex, readBytes);
    }
  }

  @MethodPattern("java.io.FileInputStream:readBytes(byte[], int, int)")
  public static final class FileInputStream_readBytes_Probe {
    public static int onEnter(final @This FileInputStream fileInputStream) {
      return processReadEnterImpl(fileInputStream);
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue int bytesRead
    ) {
      processReadReturnImpl(readRowIndex, bytesRead);
    }
  }

  private static int processReadEnterImpl(final FileInputStream fileInputStream) {
    final int streamRowIndex = ourFileInputStream2RowIndex.get(fileInputStream);
    if (streamRowIndex == Table.NO_ROW) {
      return -1;
    }

    // log individual read event
    final int readRowIndex = T_READ.createRow();
    T_READ.myInputStream.setValue(readRowIndex, streamRowIndex);
    return readRowIndex;
  }

  private static void processReadReturnImpl(final int readRowIndex, final int bytesRead) {
    if (readRowIndex == -1) {
      return;
    }

    if (bytesRead >= 0) {
      T_READ.myBytes.setValue(readRowIndex, bytesRead);
    }
    T_READ.closeRow(readRowIndex);
  }

  public static FileChannels.ChannelHandler registerInputChannel(final FileInputStream fileInputStream) {
    final int rowIndex = ourFileInputStream2RowIndex.get(fileInputStream);
    if (rowIndex == Table.NO_ROW) {
      return null;
    }
    return new MyChannelHandler(rowIndex);
  }

  public static FileChannels.ChannelHandler registerOutputChannel(final FileOutputStream fileOutputStream) {
    final int rowIndex = ourFileOutputStream2RowIndex.get(fileOutputStream);
    if (rowIndex == Table.NO_ROW) {
      return null;
    }
    return new MyChannelHandler(rowIndex);
  }

  static class MyChannelHandler implements FileChannels.ChannelHandler {
    final int myFileRowIndex;

    private static final class ChannelReadTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_INPUT_STREAM);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public ChannelReadTable() {
        super("Channel Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final ChannelReadTable T_CHANNEL_READ = new ChannelReadTable();

    private static final class ChannelWriteTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_OUTPUT_STREAM);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public ChannelWriteTable() {
        super("Channel Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final ChannelWriteTable T_CHANNEL_WRITE = new ChannelWriteTable();

    private static final class TransferFromTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_OUTPUT_STREAM);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public TransferFromTable() {
        super("Transfers from Channel", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final TransferFromTable T_TRANSFER_FROM_CHANNEL = new TransferFromTable();

    private static final class TransferToTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_INPUT_STREAM);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public TransferToTable() {
        super("Transfers to Channel", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final TransferToTable T_TRANSFER_TO_CHANNEL = new TransferToTable();

    private MyChannelHandler(final int parentRowIndex) {
      myFileRowIndex = parentRowIndex;
    }

    public int readOnEnter() {
      final int readRowIndex = T_CHANNEL_READ.createRow();
      T_CHANNEL_READ.myFile.setValue(readRowIndex, myFileRowIndex);
      return readRowIndex;
    }

    public void readOnReturn(final int readRowIndex, final long bytesRead) {
      T_CHANNEL_READ.myBytes.setValue(readRowIndex, bytesRead);
      T_CHANNEL_READ.closeRow(readRowIndex);
    }

    public int writeOnEnter() {
      final int writeRowIndex = T_CHANNEL_WRITE.createRow();
      T_CHANNEL_WRITE.myFile.setValue(writeRowIndex, myFileRowIndex);
      return writeRowIndex;
    }

    public void writeOnReturn(final int writeRowIndex, final long bytesWritten) {
      if (bytesWritten >= 0) {
        T_CHANNEL_WRITE.myBytes.setValue(writeRowIndex, bytesWritten);
      }
      T_CHANNEL_WRITE.closeRow(writeRowIndex);
    }

    public int transferToOnEnter() {
      final int transferToRowIndex = T_TRANSFER_TO_CHANNEL.createRow();
      T_TRANSFER_TO_CHANNEL.myFile.setValue(transferToRowIndex, myFileRowIndex);
      return transferToRowIndex;
    }

    public void transferToOnReturn(final int transferToRowIndex, final long transferredBytes) {
      T_TRANSFER_TO_CHANNEL.myBytes.setValue(transferToRowIndex, transferredBytes);
      T_TRANSFER_TO_CHANNEL.closeRow(transferToRowIndex);
    }

    public int transferFromOnEnter() {
      final int transferFromRowIndex = T_TRANSFER_FROM_CHANNEL.createRow();
      T_TRANSFER_FROM_CHANNEL.myFile.setValue(transferFromRowIndex, myFileRowIndex);
      return transferFromRowIndex;
    }

    public void transferFromOnReturn(final int transferFromRowIndex, final long transferredBytes) {
      T_TRANSFER_FROM_CHANNEL.myBytes.setValue(transferFromRowIndex, transferredBytes);
      T_TRANSFER_FROM_CHANNEL.closeRow(transferFromRowIndex);
    }
  }
}
