package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

public final class JUnitTests {
  private static int ourRunningTestRowIndex = Table.NO_ROW;

  private static final class TestsTable extends Table {
    private final ClassNameColumn myTestClass = new ClassNameColumn("Test Class");
    private final StringColumn myTestName = new StringColumn("Test Name");
    private final StringColumn myStatus = new StringColumn("Status");
    private final StringColumn myException = new StringColumn("Exception");

    public TestsTable() {
      super(
        "JUnit Tests",
        // test setup, run and tearDown happen in same thread
        Table.MASK_FOR_GENERIC_LASTING_EVENTS & ~Table.RECORD_THREAD_ON_ROW_CLOSE
      );
    }
  }
  private static final TestsTable T_TESTS = new TestsTable();

  private static final class TestSetupTable extends Table {
    private final ForeignKeyColumn myTest = new ForeignKeyColumn(T_TESTS);

    public TestSetupTable() {
      super("setUp", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final TestSetupTable T_SETUP = new TestSetupTable();

  private static final class TestRunTable extends Table {
    private final ForeignKeyColumn myTest = new ForeignKeyColumn(T_TESTS);

    public TestRunTable() {
      super("run", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final TestRunTable T_RUN = new TestRunTable();

  private static final class TestTearDownTable extends Table {
    private final ForeignKeyColumn myTest = new ForeignKeyColumn(T_TESTS);

    public TestTearDownTable() {
      super("tearDown", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final TestTearDownTable T_TEARDOWN = new TestTearDownTable();

  // JUnit 3.8 and older: test class must extend TestCase

  @MethodPattern("*Test:setUp()")
  public static final class OldTestCase_Before_Probe {
    public static int onEnter(final @ClassRef Class testClass) {
      if (!isTestCaseClass(testClass)) {
        return Table.NO_ROW;
      }

      return NewTestCase_Before_Probe.onEnter(testClass);
    }

    public static void onReturn(
      final @OnEnterResult int setupRowIndex
    ) {
      NewTestCase_Before_Probe.onReturn(setupRowIndex);
    }

    public static void onUncaughtException(
      final @OnEnterResult int setupRowIndex,
      final @ThrownException Throwable exception
    ) {
      NewTestCase_Before_Probe.onUncaughtException(setupRowIndex, exception);
    }
  }

  @MethodPattern("*Test:test*()")
  public static final class OldTestCase_Test_Probe {
    public static int onEnter(
      final @ClassRef Class testClass,
      final @MethodName String methodName
    ) {
      if (!isTestCaseClass(testClass)) {
        return Table.NO_ROW;
      }

      return NewTestCase_Test_Probe.onEnter(testClass, methodName);
    }

    public static void onReturn(
      final @ClassRef Class testClass,
      final @OnEnterResult int runRowIndex
    ) {
      if (runRowIndex == Table.NO_ROW) {
        return;
      }

      T_RUN.closeRow(runRowIndex);

      if (!hasTearDownMethod(testClass)) {
        closeTestRow(null);
      }
    }

    public static void onUncaughtException(
      final @ClassRef Class testClass,
      final @OnEnterResult int runRowIndex,
      final @ThrownException Throwable exception
    ) {
      if (runRowIndex == Table.NO_ROW) {
        return;
      }

      T_RUN.closeRow(runRowIndex);

      if (!hasTearDownMethod(testClass)) {
        closeTestRow(exception);
      }
    }
  }

  @MethodPattern("*Test:tearDown()")
  public static final class OldTestCase_After_Probe {
    public static int onEnter(final @ClassRef Class testClass) {
      if (!isTestCaseClass(testClass)) {
        return Table.NO_ROW;
      }

      return NewTestCase_After_Probe.onEnter();
    }

    public static void onReturn(
      final @OnEnterResult int tearDownRowIndex
    ) {
      NewTestCase_After_Probe.onReturn(tearDownRowIndex);
    }

    public static void onUncaughtException(
      final @OnEnterResult int tearDownRowIndex,
      final @ThrownException Throwable exception
    ) {
      NewTestCase_After_Probe.onUncaughtException(tearDownRowIndex, exception);
    }
  }

  private static boolean isTestCaseClass(final Class testClass) {
    try {
      if (testClass.getName().startsWith("junit.framework.")) {
        // do nothing with JUnit internals
        return false;
      }

      for (Class aClass = testClass;aClass != null; aClass = aClass.getSuperclass()) {
        if (aClass.getName().equals("junit.framework.TestCase")) {
          return true;
        }
      }

      return false;
    }
    catch (Throwable e) {
      return false;
    }
  }

  /**
   * JUnit 3.8 and older: check whether the class has overridden TestCase's tearDown()
   */
  private static boolean hasTearDownMethod(final Class testClass) {
    for (
      Class aClass = testClass;
      aClass != null && !aClass.getName().startsWith("junit.framework.");
      aClass = aClass.getSuperclass()
    ) {
      try {
        aClass.getDeclaredMethod("tearDown");
        return true;
      }
      catch (Exception ignored) {
      }
    }

    return false;
  }

  // JUnit 4.0 and newer: tests are annotated with @Before, @Test and @After

  @MethodPattern("*Test:@org.junit.Before *()")
  public static final class NewTestCase_Before_Probe {
    public static int onEnter(
      final @ClassRef Class testClass
    ) {
      ourRunningTestRowIndex = createTestRow(testClass);

      final int setupRowIndex = T_SETUP.createRow();
      T_SETUP.myTest.setValue(setupRowIndex, ourRunningTestRowIndex);

      return setupRowIndex;
    }

    public static void onReturn(
      final @OnEnterResult int setupRowIndex
    ) {
      if (setupRowIndex == Table.NO_ROW) {
        return;
      }

      T_SETUP.closeRow(setupRowIndex);
    }

    public static void onUncaughtException(
      final @OnEnterResult int setupRowIndex,
      final @ThrownException Throwable exception
    ) {
      if (setupRowIndex == Table.NO_ROW) {
        return;
      }

      T_SETUP.closeRow(setupRowIndex);

      closeTestRow(exception);
    }
  }

  @MethodPattern("*Test:@org.junit.Test *()")
  public static final class NewTestCase_Test_Probe {
    public static int onEnter(
      final @ClassRef Class testClass,
      final @MethodName String methodName
    ) {
      if (ourRunningTestRowIndex == Table.NO_ROW) {
        ourRunningTestRowIndex = createTestRow(testClass);
      }
      T_TESTS.myTestName.setValue(ourRunningTestRowIndex, methodName);

      final int runRowIndex = T_RUN.createRow();
      T_RUN.myTest.setValue(runRowIndex, ourRunningTestRowIndex);

      return runRowIndex;
    }

    public static void onReturn(
      final @ClassRef Class testClass,
      final @OnEnterResult int runRowIndex
    ) {
      if (runRowIndex == Table.NO_ROW) {
        return;
      }

      T_RUN.closeRow(runRowIndex);

      if (!hasAfterMethod(testClass)) {
        closeTestRow(null);
      }
    }

    public static void onUncaughtException(
      final @ClassRef Class testClass,
      final @OnEnterResult int runRowIndex,
      final @ThrownException Throwable exception
    ) {
      if (runRowIndex == Table.NO_ROW) {
        return;
      }

      T_RUN.closeRow(runRowIndex);

      if (!hasAfterMethod(testClass)) {
        closeTestRow(exception);
      }
    }
  }

  @MethodPattern("*Test:@org.junit.After *()")
  public static final class NewTestCase_After_Probe {
    public static int onEnter() {
      if (ourRunningTestRowIndex == Table.NO_ROW) {
        return Table.NO_ROW;
      }

      final int tearDownRowIndex = T_TEARDOWN.createRow();
      T_TEARDOWN.myTest.setValue(tearDownRowIndex, ourRunningTestRowIndex);

      return tearDownRowIndex;
    }

    public static void onReturn(
      final @OnEnterResult int tearDownRowIndex
    ) {
      if (tearDownRowIndex == Table.NO_ROW) {
        return;
      }

      T_TEARDOWN.closeRow(tearDownRowIndex);

      closeTestRow(null);
    }

    public static void onUncaughtException(
      final @OnEnterResult int tearDownRowIndex,
      final @ThrownException Throwable exception
    ) {
      if (tearDownRowIndex == Table.NO_ROW) {
        return;
      }

      T_TEARDOWN.closeRow(tearDownRowIndex);

      closeTestRow(exception);
    }
  }

  private static String getExceptionStackTrace(final Throwable e) {
    try {
      final ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
      final PrintStream printStream = new PrintStream(byteArrayOutputStream, true, "utf-8");
      try {
        e.printStackTrace(printStream);
      }
      finally {
        printStream.close();
      }

      return byteArrayOutputStream.toString("utf-8");
    }
    catch (final IOException e1) {
      // should never happen
      return e.getMessage();
    }
  }


  private static int createTestRow(final Class aClass) {
    final int rowIndex = T_TESTS.createRow();
    T_TESTS.myTestClass.setValue(rowIndex, aClass);

    return rowIndex;
  }

  /**
   * @param exception {@code null} if success
   */
  private static void closeTestRow(final Throwable exception) {
    final int rowIndex = ourRunningTestRowIndex;
    if (rowIndex == Table.NO_ROW) {
      return;
    }

    try {
      if (exception == null) {
        T_TESTS.myStatus.setValue(rowIndex, "OK");
      }
      else {
        T_TESTS.myStatus.setValue(rowIndex, "FAILED");
        T_TESTS.myException.setValue(rowIndex, getExceptionStackTrace(exception));
      }

      T_TESTS.closeRow(rowIndex);
    }
    finally {
      ourRunningTestRowIndex = Table.NO_ROW;
    }
  }

  /**
   * In new test cases: check whether the class has method annotated with @After
   */
  private static boolean hasAfterMethod(final Class testClass) {
    try {
      for (Method method : testClass.getDeclaredMethods()) {
        method.setAccessible(true);

        for (Annotation annotation : method.getAnnotations()) {
          final Class<? extends Annotation> annotationClass = annotation.annotationType();
          final String name = annotationClass.getName();
          if ("org.junit.After".equals(name)) {
            return true;
          }
        }
      }
    }
    catch (Throwable ignored) {
    }

    return false;
  }
}
