package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketAddress;
import java.nio.channels.SocketChannel;

public final class Sockets {

  /**
   * Mapping between tracked object (socket, stream or channel) to row index in T_SOCKET table
   */
  private static final ObjectRowIndexMap<Object> ourObject2SocketRowIndex = new ObjectRowIndexMap<Object>();

  private static final class SocketTable extends Table {
    private final StringColumn myAddress = new StringColumn("Address");

    public SocketTable() {
      super("Socket Connections", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final SocketTable T_SOCKET = new SocketTable();

  private static final class StreamWriteTable extends Table {
    private final ForeignKeyColumn mySocket = new ForeignKeyColumn(T_SOCKET);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public StreamWriteTable() {
      super("Stream Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final StreamWriteTable T_STREAM_WRITE = new StreamWriteTable();

  private static final class StreamReadTable extends Table {
    private final ForeignKeyColumn mySocket = new ForeignKeyColumn(T_SOCKET);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public StreamReadTable() {
      super("Stream Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final StreamReadTable T_STREAM_READ = new StreamReadTable();

  private static final class ChannelWriteTable extends Table {
    private final ForeignKeyColumn mySocketChannel = new ForeignKeyColumn(Sockets.T_SOCKET);
    private final LongColumn myBytes = new LongColumn("Bytes");

    public ChannelWriteTable() {
      super("Channel Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final ChannelWriteTable T_CHANNEL_WRITE = new ChannelWriteTable();

  private static final class ChannelReadTable extends Table {
    private final ForeignKeyColumn mySocketChannel = new ForeignKeyColumn(Sockets.T_SOCKET);
    private final LongColumn myBytes = new LongColumn("Bytes");

    public ChannelReadTable() {
      super("Channel Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final ChannelReadTable T_CHANNEL_READ = new ChannelReadTable();

  @MethodPattern("java.net.Socket:close()")
  public static final class Socket_close_Probe {
    public static void onReturn(final @This Socket socket) {
      final int rowIndex = ourObject2SocketRowIndex.remove(socket);

      if (rowIndex != Table.NO_ROW) {
        T_SOCKET.closeRow(rowIndex);
      }
    }
  }

  /**
   * Associate input streams with socket
   */
  @MethodPattern("java.net.Socket:getInputStream()")
  public static final class Socket_getInputStream_Probe {
    public static void onReturn(
      final @ReturnValue InputStream stream,
      final @This Socket socket
    ) {
      final int rowIndex = ourObject2SocketRowIndex.get(socket);
      if (rowIndex != Table.NO_ROW) {
        ourObject2SocketRowIndex.put(stream, rowIndex);
      }
    }
  }

  /**
   * Associate output streams with socket
   */
  @MethodPattern("java.net.Socket:getOutputStream()")
  public static final class Socket_getOutputStream_Probe {
    public static void onReturn(
      final @ReturnValue OutputStream stream,
      final @This Socket socket
    ) {
      final int rowIndex = ourObject2SocketRowIndex.get(socket);
      if (rowIndex != Table.NO_ROW) {
        ourObject2SocketRowIndex.put(stream, rowIndex);
      }
    }
  }

  /**
   * Log individual read event
   */
  @MethodPattern(
    "java.net.SocketInputStream:read(byte[], int, int)"
  )
  public static final class SocketInputStream_read_Probe {
    public static int onEnter(final @This InputStream stream) {
      final int socketRowIndex = ourObject2SocketRowIndex.get(stream);
      if (socketRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int readRowIndex = T_STREAM_READ.createRow();
      T_STREAM_READ.mySocket.setValue(readRowIndex, socketRowIndex);
      return readRowIndex;
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue int bytesRead
    ) {
      if (readRowIndex == -1) {
        return;
      }

      if (bytesRead >= 0) { // not EOF
        T_STREAM_READ.myBytes.setValue(readRowIndex, bytesRead);
      }
      T_STREAM_READ.closeRow(readRowIndex);
    }
  }

  @MethodPattern("java.net.ServerSocket:accept()")
  public static final class ServerSocket_accept_Probe {
    public static void onReturn(
      final @ReturnValue Socket socket
    ) {
      final SocketAddress address = socket.getLocalSocketAddress();
      if (address == null) {
        return;
      }

      final int rowIndex = T_SOCKET.createRow();
      T_SOCKET.myAddress.setValue(rowIndex, address.toString());

      ourObject2SocketRowIndex.put(socket, rowIndex);
    }
  }

  @MethodPattern("java.net.Socket:connect(java.net.SocketAddress, int)")
  public static final class Socket_connect_Probe {
    public static void onReturn(
      final @This Socket socket,
      final @Param(1) SocketAddress socketAddress
    ) {
      final int rowIndex = T_SOCKET.createRow();
      T_SOCKET.myAddress.setValue(rowIndex, socketAddress.toString());

      ourObject2SocketRowIndex.put(socket, rowIndex);
    }
  }

  /**
   * Log individual write event
   */
  @MethodPattern("java.net.SocketOutputStream:socketWrite(byte[], int, int)")
  public static final class SocketOutputStream_write_Probe {
    public static int onEnter(
      final @This OutputStream stream,
      final @Param(3) int writtenBytes
    ) {
      final int socketRowIndex = ourObject2SocketRowIndex.get(stream);
      if (socketRowIndex == Table.NO_ROW) {
        return -1;
      }

      final int rowIndex = T_STREAM_WRITE.createRow();
      T_STREAM_WRITE.myBytes.setValue(rowIndex, writtenBytes);
      T_STREAM_WRITE.mySocket.setValue(rowIndex, socketRowIndex);
      return rowIndex;
    }

    public static void onReturn(final @OnEnterResult int rowIndex) {
      if (rowIndex != -1) {
        T_STREAM_WRITE.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern("sun.nio.ch.ServerSocketChannelImpl:accept()")
  public static final class ServerSocketChannel_accept_Probe {
    public static void onReturn(final @ReturnValue SocketChannel socketChannel) {
      if (socketChannel == null) {
        return;
      }

      final Socket socket = socketChannel.socket();
      if (socket == null) {
        return;
      }

      final SocketAddress address = socket.getLocalSocketAddress();
      if (address == null) {
        return;
      }

      final int rowIndex = Sockets.T_SOCKET.createRow();
      Sockets.T_SOCKET.myAddress.setValue(rowIndex, address.toString());

      ourObject2SocketRowIndex.put(socketChannel, rowIndex);
    }
  }

  @MethodPattern("sun.nio.ch.SocketChannelImpl:connect(java.net.SocketAddress)")
  public static final class SocketChannel_connect_Probe {
    public static void onEnter(
      final @This SocketChannel socketChannel,
      final @Param(1) SocketAddress address
    ) {
      final int rowIndex = Sockets.T_SOCKET.createRow();
      Sockets.T_SOCKET.myAddress.setValue(rowIndex, address.toString());

      ourObject2SocketRowIndex.put(socketChannel, rowIndex);
    }
  }

  @MethodPattern("sun.nio.ch.SocketChannelImpl:implCloseSelectableChannel()")
  public static final class SocketChannel_close_Probe {
    public static void onReturn(final @This SocketChannel socketChannel) {
      final int rowIndex = ourObject2SocketRowIndex.remove(socketChannel);

      if (rowIndex != Table.NO_ROW) {
        Sockets.T_SOCKET.closeRow(rowIndex);
      }
    }
  }

  @MethodPattern(
    {
      "sun.nio.ch.SocketChannelImpl:write(java.nio.ByteBuffer)",
      "sun.nio.ch.SocketChannelImpl:write(java.nio.ByteBuffer[], int, int)"
    }
  )
  public static final class SocketChannel_write_Probe {
    public static int onEnter(final @This SocketChannel socketChannel) {
      final int socketChannelRowIndex = ourObject2SocketRowIndex.get(socketChannel);
      if (socketChannelRowIndex == Table.NO_ROW) {
        return -1;
      }

      // log individual write event
      final int writeRowIndex = T_CHANNEL_WRITE.createRow();
      T_CHANNEL_WRITE.mySocketChannel.setValue(writeRowIndex, socketChannelRowIndex);
      return writeRowIndex;
    }

    public static void onReturn(
      final @OnEnterResult int writeRowIndex,
      final @ReturnValue Object returnValue
    ) {
      if (writeRowIndex == -1) {
        return;
      }

      final long bytesWritten = ((Number)returnValue).longValue();
      if (bytesWritten >= 0) {
        T_CHANNEL_WRITE.myBytes.setValue(writeRowIndex, bytesWritten);
      }
      T_CHANNEL_WRITE.closeRow(writeRowIndex);
    }
  }

  @MethodPattern(
    {
      "sun.nio.ch.SocketChannelImpl:read(java.nio.ByteBuffer)",
      "sun.nio.ch.SocketChannelImpl:read(java.nio.ByteBuffer[], int, int)"
    }
  )
  public static final class SocketChannel_read_Probe {
    public static int onEnter(final @This SocketChannel socketChannel) {
      final int socketChannelRowIndex = ourObject2SocketRowIndex.get(socketChannel);
      if (socketChannelRowIndex == Table.NO_ROW) {
        return -1;
      }

      // log individual read event
      final int readRowIndex = T_CHANNEL_READ.createRow();
      T_CHANNEL_READ.mySocketChannel.setValue(readRowIndex, socketChannelRowIndex);
      return readRowIndex;
    }

    public static void onReturn(
      final @OnEnterResult int readRowIndex,
      final @ReturnValue Object returnValue
    ) {
      if (readRowIndex == -1) {
        return;
      }

      final long bytesRead = ((Number)returnValue).longValue();
      if (bytesRead >= 0) {
        T_CHANNEL_READ.myBytes.setValue(readRowIndex, bytesRead);
      }
      T_CHANNEL_READ.closeRow(readRowIndex);
    }
  }
}
